#!/bin/ksh
#set -x

PATCH_TYPE="fix pack"
PATCH_NAME="FP410T-03"
README=${PATCH_NAME}Readme.pdf
IDSRELEASE="4.1"
idsrelease="4\.1"
IDS_EXE="slapd"
RESTART_MSG=""


myid=$(id | cut -f 2 -d '(' | cut -f 1 -d ')')
if [[ ${myid} != "root" ]]; then
   print "You must be root to install this ${PATCH_TYPE}."
   return 1
fi

if ! ( rpm -qa | grep ${idsrelease} | grep ldap-client >/dev/null ); then
   print "\nThis ${PATCH_TYPE} requires IDS v${IDSRELEASE}"
   return 1
fi

if [[ -n ${IDS_EXE} ]]; then
    if [[ ${IDS_EXE} = "ibmslapd" ]]; then
        RESTART_MSG="You can restart ibmdiradm and ibmslapd."
        if ( ps -ef | grep ibmdiradm | grep -v grep ) ; then
           print "ibmdiradm must be stopped before installing this ${PATCH_TYPE}."
           return 1
        fi
    else
        RESTART_MSG="You can restart slapd."
    fi
    
    if ( ps -ef | grep ${IDS_EXE} | grep -v grep ) ; then
       print "${IDS_EXE} must be stopped before installing this ${PATCH_TYPE}."
       return 1
    fi
fi

PATCH_DIR=$(dirname ${0})
cd ${PATCH_DIR}
PATCH_DIR=$(pwd)

if [[ ! -a data/all_files.list ]]; then
   print "Error: can't find ${PATCH_DIR}/data/all_files.list"
   return 1
fi

if [[ ! -a data/UNDO.tar.gz ]]; then
   ( rm data/UNDO_files.* 2>&1 ) >/dev/null
   if [[ ! -a data/UNDO_files.list ]]; then
      for filename in $(cat data/all_files.list); do
         if [[ -f /usr/ldap/${filename} ]]; then
            print ${filename} >>data/UNDO_files.list
            cksum /usr/ldap/${filename} >>data/UNDO_files.sum
         fi
      done
   fi

   print "Backing up existing files..."
   cd /usr/ldap
   tar cf ${PATCH_DIR}/data/UNDO.tar $(cat ${PATCH_DIR}/data/UNDO_files.list)
   if [[ $? -ne 0 ]]; then
      print "\nUnable to backup existing files."
      print "The ${PATCH_TYPE} will not be installed."
      return 1
   fi
   cd ${PATCH_DIR}
   gzip data/UNDO.tar
fi

if [[ ! -a data/UNDO_files.list ]]; then
   # This should never happen, but just in case...
   print "Regenerating ${PATCH_TYPE} file list..."
   gzip -cd data/UNDO.tar.gz | tar tf - >data/UNDO_files.list
   if [[ $? -ne 0 ]]; then
      print "\nUnable to generate filelist."
      print "the ${PATCH_TYPE} will not be installed."
      return 1
   fi
fi

print "Installing ${PATCH_TYPE} files ..."
( rm -rf tmp 2>&1 ) >/dev/null
mkdir tmp
cd tmp

gzip -cd ../data/patch.tar.gz | tar xf - bin/${README} bin/${PATCH_NAME}.txt $(cat ../data/UNDO_files.list)
if [[ $? -ne 0 ]]; then
   print "\nError extracting ${PATCH_TYPE}"
   cd ..
   rm -rf tmp
   return 1
fi

## Restore backup files if copy fails
/bin/cp -r * /usr/ldap
if [[ $? -ne 0 ]]; then
   print "The ${PATCH_TYPE} could not be installed."
   print "Restoring backup files ..."
   cd /usr/ldap
   ( rm -rf bin/${PATCH_NAME}* 2>&1 ) >/dev/null
   ## Prompt user any errors if restoring fails
   gzip -cd ${PATCH_DIR}/data/UNDO.tar.gz | tar xf -
   if [[ $? -ne 0 ]]; then
      print "Error restoring files."
      print "Please manually restore the backup files from ${PATCH_DIR}/data/UNDO.tar.gz"
      return 1
   fi
   return 1
fi

cd ${PATCH_DIR}
rm -rf tmp

## Remove any existing PATCH_files
( rm data/PATCH_files.* 2>&1 ) >/dev/null
## Get sum and size for Readme, signature file and all the rest
for filename in bin/${README} bin/${PATCH_NAME}.txt $(cat data/UNDO_files.list); do
   cksum /usr/ldap/${filename} >>data/PATCH_files.sum
done

print "The ${PATCH_TYPE} has been successfully installed."
print ${RESTART_MSG}
print "\n(Note: The patch can be uninstalled by running uninstall_update.sh"
print "       only if data/UNDO.tar.gz is not removed. )"

return 0
