#ifndef __OS400__
static char sccsid[] = "@(#) 12 1.6.12.9 src/clients/tools/ldapsearch.c, ldap.clients, aus41ldap, 040224a 4/14/03 14:53:44";
#endif

/*
 * COMPONENT_NAME: ldap.clients
 *
 * ABSTRACT: generic program to search entries using LDAP
 *
 * ORIGINS: 202,27
 *
 * (C) COPYRIGHT International Business Machines Corp. 1997, 2001
 * All Rights Reserved
 * Licensed Materials - Property of IBM
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */ 
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 */

#if defined(__OS400__)
#pragma comment (copyright, \
   "5722SS1  (C) Copyright IBM Corp. 1997, 2001. \
   All rights reserved. \
   US Government Users Restricted Rights - \
   Use, duplication or disclosure restricted \
   by GSA ADP Schedule Contract with IBM Corp. \
   Licensed Materials-Property of IBM")
#endif /* OS400 */

#if defined( __OS400__ )
#include <QgldHeap.H>
#include <unistd.h>
#include <sys/stat.h>
#include "lber-int.h" /* xlate_local_to_utf8 */
#define isascii isprint
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <ldap.h>

#ifdef LDAP_SSL_MAX
#if defined(__OS400__)
#include <skit40/ibmgsk.h>
#else
#include <ldapssl.h>
#endif
#endif

#include <ldif.h>

#define DEFSEP  "="
#define DEFHOST  "localhost"

#if defined( WIN32 ) || defined( __OS400__ )
#define strncasecmp strnicmp
#define prefname "ldapsearch"
#else
#define prefname "/tmp/ldapsearch"
#endif

#if defined (_HP9000) || (sunos5)
#include <unistd.h>
#endif

#ifndef __OS400__ 
/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else 
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );
#endif

#ifndef _WIN32
extern int errno;
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = "";
static char *passwd_utf8; /* on all platforms other than OS/400 passwd == passwd_utf8 */
static char *base = NULL;
static char *ldaphost = DEFHOST;
static int ldapport = LDAP_PORT;
static char *sep = DEFSEP;
static char *mech = NULL;
static int hoplimit = 10;
static int verbose = 0;
static int allow_binary = 0;
static int ldif = 0;
static int vals2tmp = 0;
static int doit = 1;
static int ldapversion = LDAP_VERSION3;
static int DebugLevel = 0;
static int referrals = LDAP_OPT_ON;
static int attrsonly = 0;
static int scope = LDAP_SCOPE_SUBTREE;
static int deref = 0;
static FILE *fp = NULL;
static int timelimit = 0;
static int sizelimit = 0;
static int ssl = 0;
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_dn = NULL;
static char *charset = NULL;
static char *filtpattern = NULL;
static char **attrs = NULL;
static LDAP *ld;
#define FILTERSIZE 4096
static char *sortString = NULL;
static LDAPsortkey **sortKeyList = NULL;
static char sortCriticality = 'T';
static char pagingCriticality = 'T';
static int timeBetweenPages = 5;  /* default of five seconds between pages */
static struct berval *berValue = NULL;
static unsigned long totalCount = 0;
static unsigned long runningTotal = 0;
typedef struct _PageQuerySize {
   int pageSize;
   struct _PageQuerySize *next;
}
PageQuerySize;
static PageQuerySize *pageQuerySize = NULL, *curPageQuerySize = NULL, *newPageQuerySize = NULL;

static int manageDsa = LDAP_FALSE;
LDAPControl manageDsaIT = { "2.16.840.1.113730.3.4.2",    /* OID      */
                            {0, NULL},                    /* no value */
                            LDAP_OPT_ON                  /* critical */
                          };

LDAPControl *sortControl = NULL, *pageControl = NULL;
LDAPControl *M_controls[4] = { NULL, NULL, NULL, NULL };

char *debug_ptr = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends a search request to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s [-b basedn] [options] filter [attributes...]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "        basedn:     base dn for search\n");
   fprintf(stderr, "                    (optional if LDAP_BASEDN set in environment)\n");
   fprintf(stderr, "        filter:     LDAP search filter (RFC-1558 compliant)\n" );
   fprintf(stderr, "        attributes: whitespace-separated list of attributes to retrieve\n");
   fprintf(stderr, "                    (if no attribute list is specified, all are retrieved)\n");
   fprintf(stderr, "options:\n");
   fprintf(stderr, "    -a deref     how to dereference aliases (never, always, search, or find)\n");
   fprintf(stderr, "    -A           retrieve attribute names only (no values)\n");
   fprintf(stderr, "    -b base_dn   base dn for search; LDAP_BASEDN in environment is default\n");
   fprintf(stderr, "    -B           do not suppress printing of non-ASCII values\n");
   fprintf(stderr, "    -C charset   character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level     set debug level to 'level' in LDAP library\n");
   fprintf(stderr, "    -D dn        bind dn\n");
   fprintf(stderr, "    -e           display LDAP library version information and quit\n");
   fprintf(stderr, "    -f file      perform sequence of searches using filters in 'file'\n");
   fprintf(stderr, "                 \"%%s\" must be substituted for the filter\n");
   fprintf(stderr, "    -F sep       print 'sep' between attribute names and values\n");
   fprintf(stderr, "    -h host      LDAP server host name\n");
   fprintf(stderr, "    -K keyfile   file to use for keys\n");
   fprintf(stderr, "    -l time      time limit (in seconds) for search\n");
   fprintf(stderr, "    -L           print entries in LDIF format (-B is implied)\n");
   fprintf(stderr, "    -m mechanism perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M           manage referral objects as normal entries\n" );
   fprintf(stderr, "    -n           show what would be done but don't actually do it\n");
   fprintf(stderr, "    -N key_name  private key name to use in keyfile\n");
   fprintf(stderr, "    -o attr_type sort based on specified attribute type\n");
   fprintf(stderr, "    -O maxhops   maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port      LDAP server port number\n");
   fprintf(stderr, "    -P key_pw    keyfile password\n");
   fprintf(stderr, "    -q pagesize  query page size for paged results\n");
   fprintf(stderr, "    -R           do not automatically chase referrals\n");
   fprintf(stderr, "    -s scope     search scope (base, one, or sub)\n");
   fprintf(stderr, "    -t           write values to files in /tmp\n");
   fprintf(stderr, "    -T seconds   number of seconds to wait between pages for paged results\n");
   fprintf(stderr, "    -v           run in verbose mode\n");
   fprintf(stderr, "    -V version   LDAP protocol version (2 or 3; default is 3)\n");
#ifdef __OS400__

   fprintf(stderr, "    -w password  bind password\n");
#else

   fprintf(stderr, "    -w password  bind password or '?' for non-echoed prompt\n");
#endif

   fprintf(stderr, "    -z size      size limit (in entries) for search\n");
   fprintf(stderr, "    -Z           use a secure ldap connection (SSL)\n");
   exit(1);
}

/*
** Print a memory error message and exit.
*/
void mem_error(char *varName) {
   printf("%s: Memory allocation error occured.\n", varName);
   exit(1);
}

int is_num(char *s) {
   int rc = 0;
   if (s && (*s == '+' || *s == '-'))
      s++;
   while (s && *s) {
      if (!isdigit(*s))
         return rc;
      s++;
   }
   return 1;
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i, port = 0, stringLen = 0;
   char *options = "eZnvtRMABL?D:V:s:f:h:b:d:p:F:a:w:l:m:O:z:K:P:N:C:o:9:T:q:";
   LDAPVersion ver;
   extern char *optarg;
   extern int optind;
   char *emptyStr = (char *)strdup(" ");

   while (( i = getopt(argc, argv, options)) != EOF ) {
      switch ( i ) {
      case 'V':  /* use version 3 functions */
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Incorrect LDAP protocol version selected.\n");
            usage();
         }
         break;
      case 'O':        /* set maximum referral hop count  */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         hoplimit = atoi( optarg );
         if (hoplimit < 0) {
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         break;
      case 'm':        /* use SASL bind mechanism  */
         if ( ! ( mech = strdup ( optarg )))
            mem_error ( "mech" );
         break;
      case 'n':  /* do Not do any searches */
         doit = 0;
         break;
      case 'v':  /* verbose mode */
         ++verbose;
         break;
      case 'e':
         ldap_version(&ver);
         printf ("\nSDK Version:     \t%d\n", ver.sdk_version);
         printf ("Protocol Version:\t%d\n", ver.protocol_version);
         printf ("SDK Build Level: \t%s\n", ver.sdk_build_level);
         exit(0);
      case 'd':
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect debug level.\n");
            usage();
         }
         DebugLevel = atoi( optarg );
         break;
      case 't':  /* write attribute values to /tmp files */
         ++vals2tmp;
         break;
      case 'R':  /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'M':  /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         M_controls[0] = &manageDsaIT;
         break;
      case 'A':  /* retrieve attribute names only -- no values */
         ++attrsonly;
         break;
      case 'L':  /* print entries in LDIF format */
         ++ldif;
         break;
      case 'B':  /* allow binary values to be printed */
         ++allow_binary;
         break;
      case 's':  /* search scope */
         if ( strncasecmp( optarg, "base", 4 ) == 0 ) {
            scope = LDAP_SCOPE_BASE;
         } else if ( strncasecmp( optarg, "one", 3 ) == 0 ) {
            scope = LDAP_SCOPE_ONELEVEL;
         } else if ( strncasecmp( optarg, "sub", 3 ) == 0 ) {
            scope = LDAP_SCOPE_SUBTREE;
         } else {
            fprintf( stderr, "scope should be base, one, or sub\n" );
            usage();
         }
         break;
      case 'a':  /* set alias deref option */
         if ( strncasecmp( optarg, "never", 5 ) == 0 ) {
            deref = LDAP_DEREF_NEVER;
         } else if ( strncasecmp( optarg, "search", 5 ) == 0 ) {
            deref = LDAP_DEREF_SEARCHING;
         } else if ( strncasecmp( optarg, "find", 4 ) == 0 ) {
            deref = LDAP_DEREF_FINDING;
         } else if ( strncasecmp( optarg, "always", 6 ) == 0 ) {
            deref = LDAP_DEREF_ALWAYS;
         } else {
            fprintf( stderr, "alias deref should be never, search, find, or always\n" );
            usage();
         }
         break;
      case 'F':  /* field separator */
         if ( ! ( sep = strdup ( optarg )))
            mem_error( "sep" );
         break;
      case 'f':  /* input file */
         if (optarg[0] == '-' && optarg[1] == '\0' )
            fp = stdin;
         else if (( fp = fopen( optarg, "r" )) == NULL ) {
            perror( optarg );
            exit( 1 );
         }
         break;
      case 'h':  /* ldap host */
         if ( ! ( ldaphost = strdup ( optarg )))
            mem_error ( "ldaphost" );
         break;
      case 'b':  /* searchbase */
         if ( ! ( base = strdup ( optarg )))
            mem_error ( "base" );
         break;
      case 'D':  /* bind DN */
         if ( ! ( binddn = strdup ( optarg )))
            mem_error ( "binddn" );
         break;
      case 'p':  /* ldap port */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         ldapport = atoi( optarg );
         if (ldapport <= 0) {
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         port = 1;
         break;
      case 'w':  /* bind password */
#ifndef __OS400__

         if (optarg && optarg[0] == '?') {
            passwd = getPassword();

         } else
#endif
            if ( ! ( passwd = strdup ( optarg )))
               mem_error ( "passwd" );

         /* OS/400 needs the password in utf8 for CRAM-MD5
         *           and other sasl binds and rebinds */
         passwd_utf8 = passwd;
#ifdef __OS400__

         {
            unsigned long passwd_len = strlen( passwd ) + 1;
            if ( xlate_local_to_utf8( &passwd_utf8, &passwd_len, 0 ) != LDAP_SUCCESS ) {
               passwd_utf8 = NULL;
            }
         }
#endif
         break;
      case 'l':  /* time limit */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect time limit.\n");
            usage();
         }
         timelimit = atoi( optarg );
         if (timelimit < 0) {
            fprintf(stderr, "Incorrect time limit.\n");
            usage();
         }
         break;
      case 'z':  /* size limit */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect size limit.\n");
            usage();
         }
         sizelimit = atoi( optarg );
         if (sizelimit < 0 ) {
            fprintf(stderr, "Incorrect size limit.\n");
            usage();
         }
         break;
      case 'K':
         if ( ! ( keyfile = strdup ( optarg )))
            mem_error ( "keyfile" );
         ssl = 1 ;
         break;
      case 'P':
         if ( ! ( keyfile_pw = strdup ( optarg )))
            mem_error ( "keyfile_pw" );
         break;
      case 'Z':
         ssl = 1;
         break;
      case 'N':
         if ( ! ( keyfile_dn = strdup ( optarg )))
            mem_error ( "keyfile_dn" );
         break;
      case 'C':
         if (!(charset = strdup(optarg)))
            mem_error("charset");
         break;
      case 'o':
         /* copy key to sort string, prefix space as string delimeter */
         if (sortString == NULL) {
            stringLen = strlen(optarg) + 1; /* 1 for beginning blank */
            sortString = (char *) malloc(stringLen + 1); /* 1 for end of string */
            if (sortString == NULL)
               mem_error("sortString");
            memset(sortString, ' ', 1);
            memcpy(sortString + 1, optarg, strlen(optarg));
            memset(sortString + 1 + strlen(optarg), '\0', 1);
         } else {
            char *newSortString = (char *) realloc(sortString, (stringLen + strlen(optarg) + 1 + 1));
            sortString = newSortString;
            memset(sortString + stringLen, ' ', 1);
            memcpy(sortString + stringLen + 1, optarg, strlen(optarg));
            stringLen += strlen(optarg) + 1;
            memset(sortString + stringLen, '\0', 1);
         }
         break;
      case 'T':  /* time, in seconds, between pages of search results */
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect time interval.\n");
            usage();
         }
         timeBetweenPages = atoi( optarg );
         if (timeBetweenPages < 0) {
            fprintf(stderr, "Incorrect time interval.\n");
            usage();
         }
         break;
      case 'q':
         if (!is_num(optarg)) {
            fprintf(stderr, "Incorrect page size.\n");
            usage();
         }
         newPageQuerySize = (PageQuerySize *) malloc(sizeof(PageQuerySize));
         if (newPageQuerySize == NULL) {
            mem_error("pageQuerySize");
            break;
         }
         newPageQuerySize->pageSize = atoi(optarg);
         if (newPageQuerySize->pageSize < 0) {
            fprintf(stderr, "Incorrect page size.\n");
            usage();
         }
         newPageQuerySize->next = NULL;
         if (curPageQuerySize == NULL) {
            pageQuerySize = curPageQuerySize = newPageQuerySize;
         } else {
            curPageQuerySize->next = newPageQuerySize;
            curPageQuerySize = newPageQuerySize;
         }
         break;
      case '9':
         if ((strcmp(optarg, "S") == 0) || (strcmp(optarg, "s") == 0))
            sortCriticality = 'F';
         else if ((strcmp(optarg, "P") == 0) || (strcmp(optarg, "p") == 0))
            pagingCriticality = 'F';
         else {
            fprintf(stderr, "\nCriticality, option -9, is true by default.\n");
            fprintf(stderr, "Use -9 S or -9 s to set it to false for sort, and/or use -9 P or -9 p\n");
            fprintf(stderr, "to set it to false for paging.\n" );
            usage();
         }
         break;
      case '?':
      default:
         usage();
      }
   }

   if ( manageDsa && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M option requires version 3. -M ignored.\n\n");
      manageDsa = LDAP_FALSE;
   }

   if (argc - optind < 1)
      usage();

   if (!(filtpattern = strdup(argv[optind])))
      mem_error ("filtpattern");

   if (argv[optind + 1] == NULL)
      attrs = NULL;
   else
      attrs = &argv[ optind + 1 ];

   if (base == NULL) {
      base = getenv( "LDAP_BASEDN" );
      if (base != NULL)
         base = strdup(base);
   }

   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT; /* Use default SSL port if none specified */

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );

   if ( emptyStr )
      free( emptyStr );
}


#ifndef __OS400__ 
/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   fputs( "Enter password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}
#endif


/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
#ifdef __OS400__  /* use translated passwd for sasl */

         if ( mech != NULL )
            *pwp = strdup( passwd_utf8 );
         else
#endif

            *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/
void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;


   /* call ldap_ssl_client_init if V3 and SSL */
   if (ssl && (ldapversion == LDAP_VERSION3)) {

      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         exit( 1 ) ;
      }
   }

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n", ldaphost, ldapport,
                   ((keyfile_dn) ? keyfile_dn : "NULL"));
         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_dn ) ;
         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed with invalid parm\n" ) ;
            perror( ldaphost ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n", ldaphost, ldapport);
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
            fprintf( stderr, "ldap_init failed with invalid parm\n" ) ;
            perror(ldaphost);
            exit(1);
         }
      }
   } else {
      if ( verbose )
         printf( "ldap_open( %s, %d )\n", ldaphost, ldapport );
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *)&timelimit);
      ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *)&sizelimit);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )deref);
      ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *)timelimit);
      ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *)sizelimit);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)hoplimit);
   }
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )referrals);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }

   /* call ldap_ssl_start if V2 and SSL */
   if ( ssl && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_dn) ? keyfile_dn : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_dn);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         ldap_perror( ld, "ldap_ssl_start" );
         exit( 1 );
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {
      if ( ! mech ) /* Use simple bind */
      {
         rc = ldap_simple_bind_s(ld, binddn, passwd);
         if ( rc != LDAP_SUCCESS ) {
            ldap_perror( ld, "ldap_simple_bind" );
            /*
            ** LDAP_OPT_EXT_ERROR only valuable for ssl communication.
            ** In this example, for LDAP v3, the bind is the first 
            ** instance in which communication actually flows to the
            ** server.  So, if there is an ssl configuration error or
            ** other ssl problem, this will be the first instance where
            ** it will be detected.
            */
            if (ssl) {
               ldap_get_option( ld, LDAP_OPT_EXT_ERROR, &failureReasonCode);
               fprintf( stderr, "Attempted communication over SSL.\n");
               fprintf( stderr, "  The extended error is %d.\n", failureReasonCode);
            }
            exit( rc );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               ldap_perror ( ld, "ldap_sasl_bind_s" );
               exit ( rc );
            }
         } else {
            if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
               rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
               if (rc != LDAP_SUCCESS ) {
                  ldap_perror ( ld, "ldap_sasl_bind_s" );
                  exit ( rc );
               }
            } else /* other SASL mechanisms */ {
#ifdef __OS400__
               if ( passwd_utf8 == NULL ) {
                  perror("Conversion of password to utf8 failed\n");
                  exit( 1 );
               }
#endif

               ber.bv_len = strlen ( passwd_utf8 );
               ber.bv_val = passwd_utf8;
               rc = ldap_sasl_bind_s (ld, binddn, mech, &ber, NULL, NULL, &server_creds);
               if (rc != LDAP_SUCCESS ) {
                  ldap_perror ( ld, "ldap_sasl_bind_s" );
                  exit ( rc );
               }
            }
         }
      }
   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         ldap_perror( ld, "ldap_bind" );
         exit( rc );
      }
   }
}

/*
** Write an ldif style value to standard output.
*/
int write_ldif_value(char *type, char *value, unsigned long vallen) {
   char *ldif;

   if ((ldif = ldif_type_and_value(type, value, (int)vallen)) == NULL)
      return -1;
   fputs(ldif, stdout);
   free(ldif);
   return 0;
}

/*
** Write an ldif value or binary value to standard output.
*/
int write_ldif_value_or_bvalue(char *type, char *value, unsigned long vallen,
                               char *bvalue, unsigned long bvallen) {
   char *ldif;

   ldif = ldif_type_and_value_or_bvalue(type, value, (int)vallen,
                                        bvalue, (int)bvallen);
   if (ldif == NULL)
      return -1;
   fputs(ldif, stdout);
   free(ldif);
   return 0;
}

#if defined(__OS400__) 
/*
** Implement mktemp() for the OS/400 
*/
char *mktemp( char *tpl ) {
   unsigned int i;
   char *s;
   struct stat st;

   /* Find start of X's */
   for (s = tpl + strlen(tpl) - 1; s >= tpl && *s == 'X'; --s)
      ;
   ++s;

   /* Look for available name */
   for (i = 1; i < 0x01000000; ++i) {
      sprintf(s, "%6.6x", i);
      if (stat(tpl, &st))
         return tpl;
   }

   /* Never found a good name, return EEXIST error. */
   errno = 17;
   return NULL;
}
#endif

/*
** Write entry to temp file.
*/
void write_tmp_file(char *attr, struct berval *bval) {
   char postfname[128], tmpfname[128];
   FILE *tmpfp = NULL;

   sprintf(postfname, "-%s-XXXXXX", attr);
   strcpy(tmpfname, prefname);
   strcat(tmpfname, postfname);

   if (! mktemp(tmpfname)) {
      perror(tmpfname);
      return ;
   }
#ifndef __OS400__
   if ((tmpfp = fopen( tmpfname, "wb")) == NULL) {
#else
   if ((tmpfp = fopen( tmpfname, "wb, o_ccsid=1208")) == NULL) {
#endif
      perror(tmpfname);
      return ;
   }
   if (fwrite(bval->bv_val, bval->bv_len, 1, tmpfp) == 0)
      perror(tmpfname);
   else if (ldif)
      write_ldif_value(attr, tmpfname, strlen(tmpfname));
   else
      printf("%s%s%s\n", attr, sep, tmpfname);
   fclose(tmpfp);
}

/*
** Output attribute information.
*/
void print_value(char *attr, struct berval *bval, char *val) {
   int notascii = 0, i;

   if (ldif)
      write_ldif_value_or_bvalue(attr, val, bval->bv_len,
                                 bval->bv_val, bval->bv_len);
   else {
      /* if charset is set then check for ascii is invalid. */
      if (val == NULL)
         notascii = 1;
      else if (charset == NULL) {
         if (strlen(val) == bval->bv_len) {
            for (i = 0; i < bval->bv_len; i++) {
               if (!isascii(val[i])) {
                  notascii = 1;
                  break;
               }
            }
         } else
            notascii = 1;
      }

      printf("%s%s", attr, sep);
      if (notascii)
         printf("%s\n", (allow_binary) ? bval->bv_val : "NOT ASCII");
      else
         printf("%s\n", val);
   }
}

/*
** Output a search result.
*/
void print_entry(LDAPMessage *entry) {
   char *attr, *dn;
   int i;
   BerElement *ber;
   struct berval **bvals;
   char **vals;

   /* write dn */
   dn = ldap_get_dn( ld, entry );
   if (ldif)
      write_ldif_value( "dn", dn, strlen( dn ));
   else
      printf( "%s\n", dn );

   ldap_memfree( dn );

   /* write attributes */
   for (attr = ldap_first_attribute( ld, entry, &ber ); attr != NULL;
         attr = ldap_next_attribute( ld, entry, ber )) {
      /* attribute name only */
      if (attrsonly) {
         if (ldif)
            write_ldif_value(attr, "", 0);
         else
            printf("%s\n", attr);

         /* attribute name and value */
      } else if ((bvals = ldap_get_values_len(ld, entry, attr)) != NULL) {
         vals = ldap_get_values( ld, entry, attr);
         for ( i = 0; bvals[i] != NULL; i++) {
            if (vals2tmp)
               write_tmp_file(attr, bvals[i]);
            else
               print_value(attr, bvals[i],
                           (vals == NULL) ? NULL : vals[i]);
         }
         ldap_value_free_len(bvals);
         ldap_value_free(vals);
      }
      ldap_memfree(attr);
   }

   if ( ber != NULL )
      ldap_ber_free ( ber );

}

/*
** Run the search. 
*/
int dosearch(char *value) {
  char filter[FILTERSIZE ], **val;
   char **referrals = NULL;
   char *matched, *errmsg;
   int msgid, rc, first, matches, references, errcode, i;
   LDAPMessage *res, *e;
   LDAPControl **serverControls = NULL;
   char *keyInError = NULL;
   unsigned long sortRC = LDAP_SUCCESS;

   /* initialize filter */
   if ( value )
      sprintf(filter, filtpattern, value);
   else
      strncpy(filter, filtpattern, FILTERSIZE - 1);

   /* verbose */
   if (verbose)
      printf( "filter is: (%s)\n", filter );
   if (!doit)
      return LDAP_SUCCESS;

   /* send search request */
   if ( manageDsa ) {
      rc = ldap_search_ext( ld,
                            base,
                            scope,
                            filter,
                            attrs,
                            attrsonly,
                            M_controls,
                            NULL, NULL,
                            sizelimit,
                            &msgid);
      if ( rc != LDAP_SUCCESS ) {
         fprintf ( stderr, "ldap_search_ext: %s\n", ldap_err2string (rc));
         return ldap_get_errno(ld);
      }
   } else {
      msgid = ldap_search( ld, base, scope, filter, attrs, attrsonly );
      if ( msgid == -1) {
         rc = ldap_get_errno(ld);
         fprintf ( stderr, "ldap_search: %s\n", ldap_err2string (rc));
         return rc;
      }
   }
   /* disable translation if we're outputing ldif format */
   if ((ldif) && (charset != NULL))
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_OFF);

   /* get and print search results */
   matches = 0;
   references = 0;
   first = 1;
   while (1) {
      rc = ldap_result( ld, msgid, LDAP_MSG_ONE, NULL, &res );

      /* search result */
      if ( rc == LDAP_RES_SEARCH_ENTRY ) {
         matches++;
         e = ldap_first_entry( ld, res );
         if ( !first )
            putchar( '\n' );
         else
            first = 0;
         print_entry(e);
         ldap_msgfree( res );

         /* search reference */
      } else if ( rc == LDAP_RES_SEARCH_REFERENCE ) {
         references++;
         ldap_parse_reference_np( ld, res, &referrals, NULL, 1 );
         if (referrals != NULL) {
            for (i = 0; referrals[i] != NULL; i++)
               fprintf(stderr, "Unfollowed search reference: %s\n",
                       referrals[i]);
            fflush(stderr);
            ldap_value_free(referrals);
            referrals = NULL;
         }

         /* done */
      } else
         break;
   }

   /* check result code */
   if ( rc == -1 ) {
      ldap_perror( ld, "ldap_result" );
      return rc;
   }
   if (ldapversion == LDAP_VERSION3) {
      if (pageQuerySize != NULL || sortControl != NULL) {
         rc = ldap_parse_result( ld, res, &errcode, &matched, &errmsg,
                                 &referrals, &serverControls, 1 );
      } else {
         rc = ldap_parse_result( ld, res, &errcode, &matched, &errmsg,
                                 &referrals, NULL, 1 );
      }
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_search: error parsing result: %d, %s\n",
                  rc, ldap_err2string( rc ) );
      } else {
         if ( errcode != LDAP_SUCCESS ) {
            rc = errcode;  /* capture return code */
            fprintf( stderr, "ldap_search: %s\n", ldap_err2string( errcode ) );

            if ( matched != NULL )
               if ( *matched != '\0' )
                  fprintf( stderr, "ldap_search: matched: %s\n", matched );

            if ( errmsg != NULL )
               if ( *errmsg != '\0' )
                  fprintf( stderr, "ldap_search: additional info: %s\n", errmsg );

         }

         if ( matched )
            ldap_memfree ( matched );

         if ( errmsg )
            ldap_memfree ( errmsg );

         if ( referrals != NULL ) {
            int i;
            for ( i = 0; referrals[i] != NULL; i++) {
               fprintf( stderr, "%s %s\n",
                        (i == 0) ? "Unfollowed referral:" :
                        "                    ",
                        referrals[i]);
            }
            ldap_value_free( referrals );
            referrals = NULL;
         }

         /* decode page control */
         if ((serverControls != NULL) && (pageQuerySize != NULL)) {
            if (berValue)
               ber_bvfree(berValue);
            berValue = NULL;
            rc = ldap_parse_page_control(ld, serverControls, &totalCount, &berValue);

            if (rc != LDAP_SUCCESS) {
               fprintf( stderr, "ldap_search: error parsing page result: %d, %s\n",
                        rc, ldap_err2string( rc ) );
            } else {
               runningTotal += matches;
               if (verbose) {
                  printf("%d total entries have been returned\n", runningTotal);
               }
            }
         }

         /* process result controls  */
         if ((serverControls != NULL) && (sortControl != NULL)) {
            /* decode sort control */
            rc = ldap_parse_sort_control(ld, serverControls, &sortRC, &keyInError);

            if (rc != LDAP_SUCCESS) {
               /* don't print error message if this is the last page of paged results */
               /* request, we don't expect any sort control in this case    */
               fprintf( stderr, "ldap_search: error parsing sort result: %d, %s\n",
                        rc, ldap_err2string( rc ) );
            } else {
               /* display results to user */
               switch (sortRC) {
               case LDAP_SUCCESS:
                  if (verbose)
                     printf("Sorted search control successful\n");
                  break;
               case LDAP_UNWILLING_TO_PERFORM:
               case LDAP_NO_SUCH_ATTRIBUTE:
               case LDAP_INAPPROPRIATE_MATCHING:
               case LDAP_ADMIN_LIMIT_EXCEEDED:
                  if (keyInError) {
                     fprintf( stderr, "Sorted search control error:  %s, %s\n",
                              ldap_err2string(sortRC), keyInError);
                  } else {
                     fprintf( stderr, "Sorted search control error:  %s\n",
                              ldap_err2string(sortRC));
                  }
                  break;
               case LDAP_SIZELIMIT_EXCEEDED:
                  fprintf( stderr, "Sorted search control error:  %s\n",
                           ldap_err2string(LDAP_ADMIN_LIMIT_EXCEEDED));
                  break;
               case LDAP_OPERATIONS_ERROR:
               case LDAP_BUSY:
               case LDAP_INSUFFICIENT_ACCESS:
               case LDAP_UNAVAILABLE_CRITICAL_EXTENSION:
               default:
                  fprintf( stderr, "Sorted search control error:  %s\n", ldap_err2string(sortRC));
                  break;
               }
            }
         }
      }
      fflush( stderr );
   } else {
      if (( rc = ldap_result2error( ld, res, 1 )) != LDAP_SUCCESS )
         ldap_perror( ld, "ldap_search" );
   }

   if ( verbose ) {
      printf( "%d matches\n", matches );
      if (references > 0)
         printf( "%d unfollowed references\n", references );
   }

   /* re-enable translation if we previously turned it off */
   if ((ldif) && (charset != NULL))
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);

   /* clean up  */
   ldap_controls_free(serverControls);
   serverControls = NULL;
   ldap_control_free(pageControl);
   pageControl = NULL;
   if (keyInError)
      free(keyInError);
   keyInError = NULL;

   return rc ;
}


int main( int argc, char **argv ) {
   char line[ FILTERSIZE ];
   int rc = LDAP_SUCCESS, i, first, ii, firstTime = 1;
   FILE * cf_fd;

   /* Strip off any path info on program name */
   if ((prog = strrchr(argv[0], '/')) != NULL)
      ++prog;
   else
      prog = argv[0];
#ifdef __OS400__

   prog = "ldapsearch";
#endif

   passwd_utf8 = passwd;

   /* parse command line arguments */
   parse_arguments(argc, argv);

   /* connect to server */
   if (doit)
      connect_to_server();

   /* verbose */
   if ( verbose ) {
      printf( "filter pattern: %s\nreturning: ", filtpattern );
      if ( attrs == NULL )
         printf( "ALL" );
      else {
         for ( i = 0; attrs[ i ] != NULL; ++i )
            printf( "%s ", attrs[ i ] );
      }
      putchar( '\n' );
   }

   /* sorted search */
   if (sortString) {
      if ((rc = ldap_create_sort_keylist(&sortKeyList, sortString)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_search: ldap_create_sort_keylist: %d, %s\n",
                  rc, ldap_err2string( rc ) );
      } else {
         if ((rc = ldap_create_sort_control(ld, sortKeyList, sortCriticality, &sortControl)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_search: ldap_create_sort_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
         } else {
            for (ii = 0; ii < 4; ii++)
               if (M_controls[ii] == NULL) {
                  M_controls[ii] = sortControl;
                  break;
               }
            manageDsa = LDAP_TRUE;
         }
         ldap_free_sort_keylist(sortKeyList);
         if (sortString)
            free (sortString);
         sortString = NULL;
      }
   }

   /* do the search */
   if (rc == LDAP_SUCCESS) {
      if (fp == NULL) {
         if (pageQuerySize) {
            curPageQuerySize = pageQuerySize;
            while (curPageQuerySize && (firstTime || (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0)))) {
               if ((rc = ldap_create_page_control(ld, curPageQuerySize->pageSize, berValue, pagingCriticality, &pageControl)) != LDAP_SUCCESS) {
                  fprintf( stderr, "ldap_search: ldap_create_page_control: %d, %s\n",
                           rc, ldap_err2string( rc ) );
               } else {
                  for (ii = 0; ii < 4; ii++)
                     if (M_controls[ii] == NULL) {
                        M_controls[ii] = pageControl;
                        break;
                     }
                  manageDsa = LDAP_TRUE;
               }
               rc = dosearch(NULL);
               if (rc)
                  break;
               if (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0))
#ifdef _WIN32

                  Sleep(timeBetweenPages*1000);
#else

                  sleep(timeBetweenPages);
#endif

               printf( "\n" );
               if (firstTime == 1)
                  firstTime = 0;
               if (curPageQuerySize->pageSize == 0)
                  break;
               if (curPageQuerySize->next != NULL)
                  curPageQuerySize = curPageQuerySize->next;
               M_controls[ii] = NULL;
            }
         } else {
            rc = dosearch(NULL);
         }
      } else {
         rc = 0;
         first = 1;
         while ( rc == 0 && fgets( line, sizeof( line ), fp ) != NULL ) {
            if (!first)
               putchar( '\n' );
            else
               first = 0;

            /* strip newlines and carriage-returns if exists */
            i = strlen(line);
            if (i > 0 && (line[i - 1] == '\r' || line[i - 1] == '\n'))
               line[--i] = '\0';

            /* skip blank lines and comment lines */
            if (*line == '\0' ||
                  *line == '\b' ||
                  *line == '\n' ||
                  *line == '\t' ||
                  *line == ' ' ||
                  strstr(line, "/*") ||
                  strstr(line, "//"))
               continue;

            /* loop through the pages for this search filter */
            if (pageQuerySize) {
               curPageQuerySize = pageQuerySize;
               while (curPageQuerySize && (firstTime || (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0)))) {
                  if ((rc = ldap_create_page_control(ld, curPageQuerySize->pageSize, berValue, pagingCriticality, &pageControl)) != LDAP_SUCCESS) {
                     fprintf( stderr, "ldap_search: ldap_create_page_control: %d, %s\n",
                              rc, ldap_err2string( rc ) );
                  } else {
                     for (ii = 0; ii < 4; ii++)
                        if (M_controls[ii] == NULL) {
                           M_controls[ii] = pageControl;
                           break;
                        }
                     manageDsa = LDAP_TRUE;
                  }
                  rc = dosearch(line);
                  if (rc)
                     break;
                  if (berValue && berValue->bv_val != NULL && (strlen(berValue->bv_val) > 0))
#ifdef _WIN32

                     Sleep(timeBetweenPages*1000);
#else

                     sleep(timeBetweenPages);
#endif

                  printf( "\n" );
                  if (firstTime == 1)
                     firstTime = 0;
                  if (curPageQuerySize->pageSize == 0)
                     break;
                  if (curPageQuerySize->next != NULL)
                     curPageQuerySize = curPageQuerySize->next;
                  M_controls[ii] = NULL;
               }
            } else {
               rc = dosearch(line);
            }
         }
         if ( fp != stdin )
            fclose( fp );
      }
   }

   /* unbind & exit */
   if (doit)
      ldap_unbind( ld );

   /* clean up */
   ldap_control_free(sortControl);
   sortControl = NULL;
   while (pageQuerySize) {
      curPageQuerySize = pageQuerySize->next;
      free(pageQuerySize);
      pageQuerySize = curPageQuerySize;
   }
   pageQuerySize = NULL;

   exit(rc);
}
